#include <v_lib.h>

typedef vboolean (V_CALL *cplx_cart_transform_t)(void_p arg,u32_t row,u32_t col,
	vboolean neg_freq,vec_t re_im[]);

typedef vboolean (V_CALL *cplx_polar_transform_t)(void_p arg,u32_t row,u32_t col,
	vboolean neg_freq,vec_t magphase[]);

typedef vboolean (V_CALL *cplx_magnitude_transform_t)(void_p arg,
	u32_t row,u32_t col,vboolean neg_freq,vec_t *magnitude);

// #include "nr.c"
#include "util.c"

/*
typedef vec_t (*conv_kernel_t)(vec_t dist);

void
matCreateConvolutionKernel(matrix_p dest,conv_kernel_t proc)
{
	
}
*/

V_LOCAL u32_t
int_log2(u32_t val)
{
	u32_t pwr = 1;
	V_ASSERT(IS_POWER_TWO(val));
	while((u32_t)(1 << pwr) < val)
		pwr++;
	return pwr;
}

V_LOCAL void
fft_pack_row(vec_p dest,vec_p re,vec_p im,u32_t len)
{
	while(len--)
	{
		*(dest++) = *(re++);
		*(dest++) = *(im++);
	}
}

V_LOCAL void
fft_unpack_row(vec_p src,vec_p re,vec_p im,u32_t len)
{
	while(len--)
	{
		*(re++) = *(src++);
		*(im++) = *(src++);
	}
}

V_LOCAL void
fft_pack_col(vec_p dest,matrix_p re,matrix_p im,u32_t col)
{
	u32_t r;
	for(r = 0;r < re->rows;r++)
	{
		*(dest++) = re->cell[r][col];
		*(dest++) = im->cell[r][col];
	}
}

V_LOCAL void
fft_unpack_col(vec_p src,matrix_p re,matrix_p im,u32_t col)
{
	u32_t r;
	for(r = 0;r < re->rows;r++)
	{
		re->cell[r][col] = *(src++);
		im->cell[r][col] = *(src++);
	}
}

void
fft2DFullForward(matrix_p re,matrix_p im)
{
	u32_t i;
	u32_t pwr = int_log2(re->cols);
	vhandle_t ft;
	vec_p tmp = (vec_p)memAlloc(2 * re->cols * sizeof(vec_t));
	ft = fftCreateContext(pwr,false);
	for(i = 0;i < re->rows;i++)
	{
		fft_pack_row(tmp,re->cell[i],im->cell[i],re->cols);
		fftForward(ft,tmp);
		fft_unpack_row(tmp,re->cell[i],im->cell[i],re->cols);
	}
	for(i = 0;i < re->cols;i++)
	{
		fft_pack_col(tmp,re,im,i);
		fftForward(ft,tmp);
		fft_unpack_col(tmp,re,im,i);
	}
	fftDestroyContext(ft);
	memFree(tmp);
}

void
fft_mult(matrix_p re_dest,matrix_p im_dest,
	matrix_p re_mult,matrix_p im_mult)
{
	vec_t re1, re2, im1, im2;
	u32_t r, c;
	V_ASSERT(re_dest->rows == re_mult->rows);
	V_ASSERT(im_dest->rows == im_mult->rows);
	V_ASSERT(re_dest->rows == im_mult->rows);
	for(r = 0;r < re_dest->rows;r++)
	{
		for(c = 0;c < re_dest->cols;c++)
		{
			re1 = re_dest->cell[r][c];
			im1 = im_dest->cell[r][c];
			re2 = re_mult->cell[r][c];
			im2 = im_mult->cell[r][c];
			re_dest->cell[r][c] = re1 * re2 - im1 * im2;
			im_dest->cell[r][c] = im1 * re2 + re1 * im2;
		}
	}
}

void
fft_div(matrix_p re_dest,matrix_p im_dest,
	matrix_p re_mult,matrix_p im_mult)
{
	vec_t re1, re2, im1, im2, rd_mag;
	u32_t r, c;
	V_ASSERT(re_dest->rows == re_mult->rows);
	V_ASSERT(im_dest->rows == im_mult->rows);
	V_ASSERT(re_dest->rows == im_mult->rows);
	for(r = 0;r < re_dest->rows;r++)
	{
		for(c = 0;c < re_dest->cols;c++)
		{
			re1 = re_dest->cell[r][c];
			im1 = im_dest->cell[r][c];
			re2 = re_mult->cell[r][c];
			im2 = im_mult->cell[r][c];
			rd_mag = re2 * re2 + im2 * im2;
			if(IsZero(rd_mag)) /* what else can we do ? */
				continue;
			rd_mag = 1 / rd_mag;
			re_dest->cell[r][c] = rd_mag * (re1 * re2 + im1 * im2);
			im_dest->cell[r][c] = rd_mag * (im1 * re2 - re1 * im2);
		}
	}
}

void save(matrix_p m)
{
	u32_t r,c;
	for(r=0;r < m->rows;r++)
	{
		for(c=0;c<m->cols;c++)
			printf("%g\t",m->cell[r][c]);
		puts("");
	}
}

V_LOCAL void
unpack_twoffts(vec_p freq_zero,vec_p freq_middle,vec_p raw,u32_t n)
{
	u32_t j;
	u32_t nn2 = n + n;
	u32_t nn3 = 1 + nn2;
	vec_t rep, rem, aip, aim;
	freq_zero[0] = raw[0];
	freq_middle[0] = raw[1];
	for(j = 2;j < n;j += 2)
	{
		rep = (vec_t)0.5 * (raw[j] + raw[nn2 - j]);
		rem = (vec_t)0.5 * (raw[j] - raw[nn2 - j]);
		aip = (vec_t)0.5 * (raw[j + 1] + raw[nn3 - j]);
		aim = (vec_t)0.5 * (raw[j + 1] - raw[nn3 - j]);
		freq_zero[j] = rep;
		freq_zero[j + 1] = aim;
		freq_middle[j] = aip;
		freq_middle[j + 1] = -rem;
	}
	rep = (vec_t)0.5 * (raw[j] + raw[nn2 - j]); /* rem == 0 */
	aip = (vec_t)0.5 * (raw[j + 1] + raw[nn3 - j]); /* aim == 0 */
	freq_zero[1] = rep;
	freq_middle[1] = aip;
}

V_LOCAL	void
pack_twoffts(vec_p raw,vec_p freq_zero,vec_p freq_middle,u32_t n)
{
	u32_t j;
	u32_t nn2 = n + n;
	u32_t nn3 = 1 + nn2;
	vec_t rep, rem, aip, aim;
	raw[0] = freq_zero[0];
	raw[1] = freq_middle[0];
	for(j = 2;j < n;j += 2)
	{
		rep = freq_zero[j];
		aim = freq_zero[j + 1];
		aip = freq_middle[j];
		rem = - freq_middle[j + 1];
		raw[j] = rep + rem;
		raw[nn2 - j] = rep - rem;
		raw[j + 1] = aip + aim;
		raw[nn3 - j] = aip - aim;
	}
	rep = freq_zero[1];
	aip = freq_middle[1];
	raw[j] = rep;
	raw[nn2 - j] = rep;
	raw[j + 1] = aip;
	raw[nn3 - j] = aip;
}

void V_CALL
rvmUnpackCplxCol(vec_p cplx,matrix_p data,u32_t col);

void V_CALL
rvmPackCplxCol(matrix_p data,vec_p cplx,u32_t col);

void
matRVMFTMultiply(matrix_p dest,matrix_p multiplier)
{
	vec_p raw, zero, middle;
	vec_p rawm, zerom, middlem;
	vec_t re1, re2, im1, im2;
	u32_t r, c;
	V_ASSERT(dest->rows == multiplier->rows);
	V_ASSERT(dest->cols == multiplier->cols);
	/*
	 * Extract and unpack zero frequencies
	 */
	raw = (vec_p)memAlloc(dest->rows * 8 * sizeof(vec_t));
	zero = raw + dest->rows * 2;
	middle = zero + dest->rows;
	rawm = middle + dest->rows;
	zerom = rawm + dest->rows * 2;
	middlem = zerom + dest->rows;
	rvmUnpackCplxCol(raw,dest,0);
	rvmUnpackCplxCol(rawm,multiplier,0);
	unpack_twoffts(zero,middle,raw,dest->rows);
	unpack_twoffts(zerom,middlem,rawm,dest->rows);
	/*
	 * Multiply first two real elements
	 */
	zero[0] *= zerom[0];
	middle[0] *= middlem[0];
	zero[1] *= zerom[1];
	middle[1] *= middlem[1];
	/*
	 * The other elements are complex
	 */
	for(c = 2;c < dest->cols;c += 2)
	{
		re1 = zero[c];
		im1 = zero[c + 1];
		re2 = zerom[c];
		im2 = zerom[c + 1];
		zero[c] = re1 * re2 - im1 * im2;
		zero[c + 1] = im1 * re2 + re1 * im2;
		re1 = middle[c];
		im1 = middle[c + 1];
		re2 = middlem[c];
		im2 = middlem[c + 1];
		middle[c] = re1 * re2 - im1 * im2;
		middle[c + 1] = im1 * re2 + re1 * im2;
	}
	/*
	 * Pack zero frequency and reinsert it in 'dest' again
	 */
	pack_twoffts(raw,zero,middle,dest->rows);
	rvmPackCplxCol(dest,raw,0);
	memFree(raw);
	/* the other is straightforward */
	for(r = 0;r < dest->rows;r++)
	{
		for(c = 2;c < dest->cols;c += 2)
		{
			re1 = dest->cell[r][c];
			im1 = dest->cell[r][c + 1];
			re2 = multiplier->cell[r][c];
			im2 = multiplier->cell[r][c + 1];
			dest->cell[r][c] = re1 * re2 - im1 * im2;
			dest->cell[r][c + 1] = im1 * re2 + re1 * im2;
		}
	}
}

#if 0

void
matRVMFTDivide(matrix_p dest,matrix_p divisor)
{
	vec_p raw, zero, middle;
	vec_p rawm, zerom, middlem;
	vec_t re1, re2, im1, im2;
	u32_t r, c;
	V_ASSERT(dest->rows == multiplier->rows);
	V_ASSERT(dest->cols == multiplier->cols);
	/*
	 * Extract and unpack zero frequencies
	 */
	raw = (vec_p)memAlloc(dest->rows * 8 * sizeof(vec_t));
	zero = raw + dest->rows * 2;
	middle = zero + dest->rows;
	rawm = middle + dest->rows;
	zerom = rawm + dest->rows * 2;
	middlem = zerom + dest->rows;
	rvmUnpackCplxCol(raw,dest,0);
	rvmUnpackCplxCol(rawm,multiplier,0);
	unpack_twoffts(zero,middle,raw,dest->rows);
	unpack_twoffts(zerom,middlem,rawm,dest->rows);
	/*
	 * Divide first two real elements
	 */
#error
	if(!IsZero(zerom[0]))
		zero[0] /= zerom[0];
	if(!IsZero(
		middle[0] *= middlem[0];
	if(!IsZero(
		zero[1] *= zerom[1];
	if(!IsZero(
		middle[1] *= middlem[1];
	/*
	 * The other elements are complex
	 */
	for(c = 2;c < dest->cols;c += 2)
	{
		re1 = zero[c];
		im1 = zero[c + 1];
		re2 = zerom[c];
		im2 = zerom[c + 1];
		zero[c] = re1 * re2 - im1 * im2;
		zero[c + 1] = im1 * re2 + re1 * im2;
		re1 = middle[c];
		im1 = middle[c + 1];
		re2 = middlem[c];
		im2 = middlem[c + 1];
		middle[c] = re1 * re2 - im1 * im2;
		middle[c + 1] = im1 * re2 + re1 * im2;
	}
	/*
	 * Pack zero frequency and reinsert it in 'dest' again
	 */
	pack_twoffts(raw,zero,middle,dest->rows);
	rvmPackCplxCol(dest,raw,0);
	memFree(raw);
	/* the other is straightforward */
	for(r = 0;r < dest->rows;r++)
	{
		for(c = 2;c < dest->cols;c += 2)
		{
			re1 = dest->cell[r][c];
			im1 = dest->cell[r][c + 1];
			re2 = multiplier->cell[r][c];
			im2 = multiplier->cell[r][c + 1];
			dest->cell[r][c] = re1 * re2 - im1 * im2;
			dest->cell[r][c + 1] = im1 * re2 + re1 * im2;
		}
	}
}

#endif

int main()
{
	matrix_p m, psf;
	vhandle_t ctx;
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);

	m = load_image("in.bmp");
	assert(m);
	psf = load_image("kernel.bmp");
	assert(psf);

/*
	psf = matAlloc(m->rows,m->cols);
	matZero(psf);
	psf->cell[0][0] = 1;
	psf->cell[100][0] = 0.5;
	psf->cell[0][100] = 0.5;
	psf->cell[100][100] = 0.3;
*/

	ctx = fftRVMCreateContext(m->rows,m->cols);
	if(ctx == LIBV_INVALID_HANDLE)
		return -1;
	fftRVMForward(ctx,m);
	fftRVMForward(ctx,psf);
	matRVMFTMultiply(m,psf);
	matZero(psf);
	psf->cell[255][255] = 1;
	fftRVMForward(ctx,psf);
	matRVMFTMultiply(m,psf);
	fftRVMInverse(ctx,m);
	cspaceMaximizeContrast(m);
	save_image("out.bmp",m);
	memFree(m);
	memFree(psf);
	fftRVMDestroyContext(ctx);
	memPrintStats();	
	return 0;
}

