matrix_p load_image(char *name)
{
	u32_t w,h,i,j;
	matrix_p m;
	vec_t scale = 1.0 / 3 / 256;
	u8_t *ptr,*img = bmpLoad(name,&w,&h);
	if(!img)
		return NULL;
	m = matAlloc(h,w);
	ptr = img;
	for(i=0;i<m->rows;i++)
		for(j=0;j<m->cols;j++)
		{
			m->cell[i][j] = (ptr[0]+ptr[1]+ptr[2]) * scale;
			ptr += 4;
		}
	memFree(img);
	return m;
}

void save_image(char *name,matrix_p src)
{
	u32_t i,j;
	vec_t val;
	u8_t *ptr, *img = (u8_p)memAllocZ(src->rows * src->cols * 4);
	if(!img)
		return;
	ptr = img;
	for(i=0;i < src->rows;i++)
		for(j=0;j < src->cols;j++)
		{
			val = src->cell[i][j] * 255;
			if(val < 0)
				val = 0;
			else if(val > 255)
				val = 255;
			ptr[0] = ptr[1] = ptr[2] = (u8_t)val;
			ptr += 4;
		}
	bmpSave(name,src->cols,src->rows,img);
	memFree(img);
}

vec_t mat_diff(matrix_p m1,matrix_p m2)
{
	u32_t i,j;
	vec_t t,d = 0;
	V_ASSERT(m1->rows==m2->rows && m1->cols==m2->cols);
	for(i=0;i<m1->rows;i++)
		for(j=0;j<m1->cols;j++)
		{
			t = v_abs(m1->cell[i][j] - m2->cell[i][j]);
			if(t > d)
				d = t;
		}
	return d;
}

void
matZeroRows(matrix_p m,u32_t row_start,u32_t row_end)
{
	u32_t r, row_bytes;
	V_ASSERT(m);
	if(row_start > row_end || row_start >= m->rows)
		return;
	if(row_end >= m->rows)
		row_end = m->rows - 1;
	row_bytes = m->cols * sizeof(vec_t);
	for(r = row_start;r <= row_end;r++)
		memZero(m->cell[r],row_bytes);
}

void
matZeroCols(matrix_p m,u32_t col_start,u32_t col_end)
{
	u32_t c, r;
	V_ASSERT(m);
	if(col_start > col_end || col_start >= m->cols)
		return;
	if(col_end >= m->cols)
		col_end = m->cols - 1;
	for(c = col_start;c <= col_end;c++)
		for(r = 0;r < m->rows;r++)
			m->cell[r][c] = 0;
}

void
matRVMFTZeroRows(matrix_p rvmft,u32_t row_start,u32_t row_end)
{
	u32_t r, ft_len, row_bytes;
	V_ASSERT(rvmft);
	V_ASSERT(row_start <= row_end);
	/* Clamp the range to positive part of Fourier transform */
	ft_len = rvmft->rows >> 1;
	if(row_end > ft_len)
		row_end = ft_len;
	row_bytes = rvmft->cols * sizeof(vec_t);
	if(!row_start)
	{
		/* Clear zero frequency */
		memZero(rvmft->cell[0],row_bytes);
		row_start++;
	}
	/*
	 * Clear both positive and negative frequencies; there
	 * is a little overhead to do this in one loop since at
	 * the highest frequency r == rvmft->rows - r.
	 */
	for(r = row_start;r <= row_end;r++)
	{
		memZero(rvmft->cell[r],row_bytes);
		memZero(rvmft->cell[rvmft->rows - r],row_bytes);
	}
}

void
matRVMFTZeroCols(matrix_p rvmft,u32_t col_start,u32_t col_end)
{
	u32_t r, ft_len, row_bytes;
	V_ASSERT(rvmft);
	V_ASSERT(col_start <= col_end);
	/* Clamp the range to positive part of Fourier transform */
	ft_len = rvmft->cols >> 1;
	if(col_end >= ft_len)
		col_end = ft_len - 1;
	if(col_start == col_end)
	{
		for(r = 0;r < rvmft->rows;r++)
		{
			rvmft->cell[r][col_start << 1] = 0;
			rvmft->cell[r][(col_start << 1) + 1] = 0;
		}
	}
	else
	{
		row_bytes = (col_end - col_start + 1) * sizeof(vec_t) * 2;
		for(r = 0;r < rvmft->rows;r++)
			memZero(&rvmft->cell[r][col_start << 1],row_bytes);
	}
}

void
matRVMFTLowPass(matrix_p rvmft,u32_t cutoff)
{
	u32_t reject_from = cutoff + 1;
	if(reject_from > (rvmft->cols >> 1))
		reject_from = (rvmft->cols >> 1);
	matZeroCols(rvmft,reject_from * 2,rvmft->cols - 1);
	matZeroRows(rvmft,reject_from,rvmft->rows - reject_from);
}

void
matRVMFTSetRect(matrix_p rvmft,
	u32_t col_start,u32_t row_start,
	u32_t col_end,u32_t row_end,
	vec_t mag,vec_t phase)
{
	vec_t im, re;
	u32_t r, nc, ftc, ftr, row_ft_len, col_ft_len;
	V_ASSERT(rvmft);
	V_ASSERT(rvmft->cols > 1);
	V_ASSERT(IS_POWER_TWO(rvmft->cols) && IS_POWER_TWO(rvmft->rows));
	V_ASSERT(row_start <= row_end);
	V_ASSERT(col_start <= col_end);
	/* Clamp the ranges to positive part of Fourier transform */
	row_ft_len = rvmft->rows >> 1;
	col_ft_len = rvmft->cols >> 1;
	if(row_end >= row_ft_len)
		row_end = row_ft_len - 1;
	if(col_end >= col_ft_len)
		col_end = col_ft_len - 1;
	re = mag * (vec_t)cos(phase);
	im = mag * (vec_t)sin(phase);
	if(!row_start)
	{
		/* Set zero frequency */
		nc = col_end - col_start + 1;
		ftc = col_start << 1;
		while(nc--)
		{
			rvmft->cell[0][ftc++] = re;
			rvmft->cell[0][ftc++] = im;
		}
		row_start++;
	}
	/* set both positive and negative frequencies */
	for(r = row_start;r <= row_end;r++)
	{
		ftr = rvmft->rows - r;
		nc = col_end - col_start + 1;
		ftc = col_start << 1;
		while(nc--)
		{
			rvmft->cell[ftr][ftc] = re;
			rvmft->cell[r][ftc++] = re;
			rvmft->cell[ftr][ftc] = im;
			rvmft->cell[r][ftc++] = im;
		}
	}
	/* set highest frequency */
}

V_LOCAL	vec_t
cplx_phase(vec_t re,vec_t im)
{
	/* reject degenerate cases */
	if(IsZero(re))
	{
		if(im < 0)
			return -M_PI2;
		return M_PI2;
	}
	return atan2(im, re);
}

void
matRVMFTCartesianTransform(matrix_p rvmft,
	u32_t col_start,u32_t row_start,
	u32_t col_end,u32_t row_end,
	cplx_cart_transform_t proc,void_p proc_arg)
{
	u32_t r, nc, ftc, ftr, row_ft_len, col_ft_len;
	V_ASSERT(rvmft);
	V_ASSERT(rvmft->cols > 1);
	V_ASSERT(IS_POWER_TWO(rvmft->cols) && IS_POWER_TWO(rvmft->rows));
	V_ASSERT(row_start <= row_end);
	V_ASSERT(col_start <= col_end);
	/* Clamp the ranges to positive part of Fourier transform */
	row_ft_len = rvmft->rows >> 1;
	col_ft_len = rvmft->cols >> 1;
	if(row_end > row_ft_len)
		row_end = row_ft_len;
	if(col_end >= col_ft_len)
		col_end = col_ft_len - 1;
	if(!row_start)
	{
		/* at zero frequency */
		nc = col_end - col_start + 1;
		ftc = col_start << 1;
		for(;nc--;ftc += 2)
			if(!proc(proc_arg,0,ftc >> 1,false,&rvmft->cell[0][ftc]))
				goto outer_break;
		row_start++;
	}
	/* at both positive and negative frequencies */
	for(r = row_start;r <= row_end;r++)
	{
		ftr = rvmft->rows - r;
		nc = col_end - col_start + 1;
		ftc = col_start << 1;
		for(;nc--;ftc += 2)
		{
			if(!proc(proc_arg,r,ftc >> 1,false,&rvmft->cell[r][ftc]))
				goto outer_break;
			if(!proc(proc_arg,r,ftc >> 1,true,&rvmft->cell[ftr][ftc]))
				goto outer_break;
		}
	}
	/* at highest frequency */
	if(r == row_ft_len)
	{
		nc = col_end - col_start + 1;
		ftc = col_start << 1;
		for(;nc--;ftc += 2)
		{
			if(!proc(proc_arg,r,ftc >> 1,false,&rvmft->cell[r][ftc]))
				goto outer_break;
		}
	}
outer_break:;
}

void
matRVMFTPolarTransform(matrix_p rvmft,
	u32_t col_start,u32_t row_start,
	u32_t col_end,u32_t row_end,
	cplx_polar_transform_t proc,void_p proc_arg)
{
	vec_t re, im, magphase[2];
	u32_t r, nc, ftc, ftr, row_ft_len, col_ft_len;
	V_ASSERT(rvmft);
	V_ASSERT(rvmft->cols > 1);
	V_ASSERT(IS_POWER_TWO(rvmft->cols) && IS_POWER_TWO(rvmft->rows));
	V_ASSERT(row_start <= row_end);
	V_ASSERT(col_start <= col_end);
	/* Clamp the ranges to positive part of Fourier transform */
	row_ft_len = rvmft->rows >> 1;
	col_ft_len = rvmft->cols >> 1;
	if(row_end > row_ft_len)
		row_end = row_ft_len;
	if(col_end >= col_ft_len)
		col_end = col_ft_len - 1;
	if(!row_start)
	{
		/* at zero frequency */
		nc = col_end - col_start + 1;
		ftc = col_start << 1;
		while(nc--)
		{
			re = rvmft->cell[0][ftc];
			im = rvmft->cell[0][ftc+1];
			magphase[0] = (vec_t)hypot(re, im);
			magphase[1] = cplx_phase(re, im);
			if(!proc(proc_arg,0,ftc >> 1,false,magphase))
				goto outer_break;
			rvmft->cell[0][ftc++] = magphase[0] * (vec_t)cos(magphase[1]);
			rvmft->cell[0][ftc++] = magphase[0] * (vec_t)sin(magphase[1]);
		}
		row_start++;
	}
	/* at both positive and negative frequencies */
	for(r = row_start;r < row_end;r++)
	{
		ftr = rvmft->rows - r;
		nc = col_end - col_start + 1;
		ftc = col_start << 1;
		for(;nc--;ftc += 2)
		{
			re = rvmft->cell[r][ftc];
			im = rvmft->cell[r][ftc+1];
			magphase[0] = (vec_t)hypot(re, im);
			magphase[1] = cplx_phase(re, im);
			if(!proc(proc_arg,r,ftc >> 1,false,magphase))
				goto outer_break;
			rvmft->cell[r][ftc] = magphase[0] * (vec_t)cos(magphase[1]);
			rvmft->cell[r][ftc+1] = magphase[0] * (vec_t)sin(magphase[1]);
			re = rvmft->cell[ftr][ftc];
			im = rvmft->cell[ftr][ftc+1];
			magphase[0] = (vec_t)hypot(re, im);
			magphase[1] = cplx_phase(re, im);
			if(!proc(proc_arg,r,ftc >> 1,true,magphase))
				goto outer_break;
			rvmft->cell[ftr][ftc] = magphase[0] * (vec_t)cos(magphase[1]);
			rvmft->cell[ftr][ftc+1] = magphase[0] * (vec_t)sin(magphase[1]);
		}
	}
	/* at highest frequency */
	if(r == row_ft_len)
	{
		nc = col_end - col_start + 1;
		ftc = col_start << 1;
		for(;nc--;ftc += 2)
		{
			re = rvmft->cell[r][ftc];
			im = rvmft->cell[r][ftc+1];
			magphase[0] = (vec_t)hypot(re, im);
			magphase[1] = cplx_phase(re, im);
			if(!proc(proc_arg,r,ftc >> 1,false,magphase))
				goto outer_break;
			rvmft->cell[r][ftc] = magphase[0] * (vec_t)cos(magphase[1]);
			rvmft->cell[r][ftc+1] = magphase[0] * (vec_t)sin(magphase[1]);
		}
	}
outer_break:;
}

void
matScaleAdd(matrix_p dest,matrix_p src,vec_t scale,vec_t offset)
{
	u32_t r, c;
	V_ASSERT(dest->rows == src->rows);
	V_ASSERT(dest->cols == src->cols);
	for(r = 0;r < dest->rows;r++)
		for(c = 0;c < dest->cols;c++)
			dest->cell[r][c] = offset + src->cell[r][c] * scale;
}

void
matBandReject(matrix_p m,u32_t lo_bound,u32_t hi_bound)
{
	matZeroCols(m,lo_bound,hi_bound);
	matZeroRows(m,lo_bound,hi_bound);
}

void
matBandReject2(matrix_p m,u32_t lo_bound,u32_t hi_bound)
{
	u32_t neg_lower, neg_upper;
	/*
	 * Assume 'm' is a result of complex Fourier transform.
	 */
	/* Zero positive frequencies. */
	matZeroCols(m,lo_bound,hi_bound);
	matZeroRows(m,lo_bound,hi_bound);
	/* Zero negative frequencies. */
	neg_lower = m->cols - hi_bound;
	neg_upper = m->cols - lo_bound;
	if(neg_upper >= m->cols)
		neg_upper = m->cols - 1;
	if(neg_lower <= neg_upper)
		matZeroCols(m,neg_lower,neg_upper);
	neg_lower = m->rows - hi_bound;
	neg_upper = m->rows - lo_bound;
	if(neg_upper >= m->rows)
		neg_upper = m->rows - 1;
	if(neg_lower <= neg_upper)
		matZeroRows(m,neg_lower,neg_upper);
}

void
matBandPass(matrix_p m,u32_t lo_bound,u32_t hi_bound)
{
	if(lo_bound)
		matBandReject(m,0,lo_bound-1);
	if(hi_bound < m->rows-1)
		matZeroRows(m,hi_bound+1,m->rows-1);
	if(hi_bound < m->cols-1)
		matZeroCols(m,hi_bound+1,m->cols-1);
}

void
matBandPass2(matrix_p m,u32_t lo_bound,u32_t hi_bound)
{
	u32_t r_limit, c_limit;
	if(m->cols <= 1 || m->rows <= 1)
		return;
	r_limit = (m->rows >> 1) - 1;
	c_limit = (m->cols >> 1) - 1;
	if(lo_bound)
		matBandReject2(m,0,lo_bound - 1);
	if(hi_bound < r_limit)
		matZeroRows(m,hi_bound + 1,m->rows - hi_bound);
	if(hi_bound < c_limit)
		matZeroCols(m,hi_bound + 1,m->cols - hi_bound);
}

vboolean
cspaceAdjustContrast(matrix_p dest,matrix_p src,vec_t min,vec_t max)
{
	vec_t rscale;
    rscale = max - min;
    if(!IsZero(rscale))
    {
    	rscale = 1 / rscale;
    	matScaleAdd(dest,src,rscale,- min * rscale);
    	return true;
    }
   	conlPuts(LOG_WARNING,"cspaceAdjustContrast(): zero input range, can't transform.\n");
    return false;
}

vboolean
cspaceMaximizeContrast(matrix_p m)
{
	u32_t r, c;
	vec_t min = 0, max = 0;
    matMin(m,&r,&c);
    min = m->cell[r][c];
    matMax(m,&r,&c);
    max = m->cell[r][c];
    return cspaceAdjustContrast(m,m,min,max);
}

matrix_p
matPSEofRVMFT(matrix_p rvmft)
{
	u32_t r, c;
	matrix_p row_pse = matAlloc(rvmft->rows >> 1,rvmft->cols >> 1);
	vec_p *cell = rvmft->cell;
	for(c = 0;c < row_pse->cols;c++)
		row_pse->cell[0][c] =
			2 * hypot(cell[0][c << 1],cell[0][(c << 1) + 1]);
	for(r = 1;r < row_pse->rows;r++)
	{
		u32_t neg_r = rvmft->rows - r;
		for(c = 0;c < row_pse->cols;c++)
			row_pse->cell[r][c] =
				hypot(cell[r][c << 1],cell[r][(c << 1) + 1]);
		for(c = 0;c < row_pse->cols;c++)
			row_pse->cell[r][c] +=
				hypot(cell[neg_r][c << 1],cell[neg_r][(c << 1) + 1]);
	}
	return row_pse;
}

void matAdd1(matrix_p m,vec_t val)
{
	u32_t r;
	vec_p ptr, endptr;
	for(r = 0;r < m->rows;r++)
	{
		ptr = m->cell[r];
		endptr = ptr + m->cols;
		for(;ptr < endptr;ptr++)
			*ptr += val;
	}
}

void matLn(matrix_p m)
{
	u32_t r;
	vec_p ptr, endptr;
	for(r = 0;r < m->rows;r++)
	{
		ptr = m->cell[r];
		endptr = ptr + m->cols;
		for(;ptr < endptr;ptr++)
			*ptr = log(*ptr);
	}
}

vboolean V_CALL
cart_trans(void_p arg,u32_t r,u32_t c,vboolean negfreq,vec_t reim[])
{
	vboolean diag = abs(r - c) < 16;
	vboolean side = r < 16 || c < 16;
	if(!diag && !side)
		reim[0] = reim[1] = 0;
	return true;
}

vboolean V_CALL
polar_trans(void_p arg,u32_t r,u32_t c,vboolean negfreq,vec_t magphase[])
{
	vboolean diag = abs(r - c) < 4;
	vboolean side = r < 4 || c < 4;
	if(diag && side)
		magphase[0] *= -2;
	else
	{
		diag = abs(r - c) < 64;
		side = r < 64 || c < 64;
		if(diag && side)
			magphase[0] *= 3;
	}
	return true;
}

